# -*- coding: utf-8 -*-
#
# gPodder - A media aggregator and podcast client
# Copyright (c) 2005-2018 The gPodder Team
#
# gPodder is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# gPodder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


#
# gpodder.gtkui.desktop.podcastdirectory - Podcast directory Gtk UI
# Thomas Perl <thp@gpodder.org>; 2014-10-22
#


import html
import logging
import os
import traceback

from gi.repository import Gdk, GdkPixbuf, Gtk, Pango

import gpodder
from gpodder import directory, util
from gpodder.gtkui.interface.common import BuilderWidget
from gpodder.gtkui.interface.progress import ProgressIndicator
from gpodder.gtkui.interface.tagcloud import TagCloud

_ = gpodder.gettext

logger = logging.getLogger(__name__)


class DirectoryPodcastsModel(Gtk.ListStore):
    C_SELECTED, C_MARKUP, C_TITLE, C_URL, C_SECTION = list(range(5))

    def __init__(self, callback_can_subscribe):
        Gtk.ListStore.__init__(self, bool, str, str, str, str)
        self.callback_can_subscribe = callback_can_subscribe

    def load(self, directory_entries):
        self.clear()
        for entry in directory_entries:
            if entry.subscribers != -1:
                subscribers_part = '(%d)' % entry.subscribers
            else:
                subscribers_part = ''
            if entry.section:
                section_part = '%s\n' % (html.escape(entry.section))
            else:
                section_part = ''
            self.append((False, '%s%s%s\n<small>%s</small>' % (section_part, html.escape(entry.title),
                subscribers_part, html.escape(entry.url)), entry.title, entry.url, entry.section))
        self.callback_can_subscribe(len(self.get_selected_podcasts()) > 0)

    def toggle(self, path):
        self[path][self.C_SELECTED] = not self[path][self.C_SELECTED]
        self.callback_can_subscribe(len(self.get_selected_podcasts()) > 0)

    def set_selection_to(self, selected):
        for row in self:
            row[self.C_SELECTED] = selected
        self.callback_can_subscribe(len(self.get_selected_podcasts()) > 0)

    def get_selected_podcasts(self):
        return [(row[self.C_TITLE], row[self.C_URL], row[self.C_SECTION]) for row in self if row[self.C_SELECTED]]


class DirectoryProvidersModel(Gtk.ListStore):
    C_WEIGHT, C_TEXT, C_ICON, C_PROVIDER = list(range(4))

    SEPARATOR = (Pango.Weight.NORMAL, '', None, None)

    def __init__(self, providers):
        Gtk.ListStore.__init__(self, int, str, GdkPixbuf.Pixbuf, object)
        for provider in providers:
            instance = provider() if provider else None
            if instance is not None and instance.kind == directory.Provider.PROVIDER_TAGCLOUD:
                logger.warning("PROVIDER_TAGCLOUD is unsupported")
            else:
                self.add_provider(instance)

    def add_provider(self, provider):
        if provider is None:
            self.append(self.SEPARATOR)
        else:
            pixbuf = None
            if provider.icon:
                search_path = {gpodder.images_folder, }
                # let an extension provide an icon by putting it next to the source code
                for e in gpodder.user_extensions.filenames:
                    search_path.add(os.path.dirname(e))
                for d in search_path:
                    path_to_try = os.path.join(d, provider.icon)
                    if os.path.exists(path_to_try):
                        try:
                            pixbuf = GdkPixbuf.Pixbuf.new_from_file(path_to_try)
                            break
                        except Exception as e:
                            logger.warning('Could not load icon: %s (%s)', provider.icon, e)
            self.append((Pango.Weight.NORMAL, provider.name, pixbuf, provider))

    def is_row_separator(self, model, it):
        return self.get_value(it, self.C_PROVIDER) is None


class gPodderPodcastDirectory(BuilderWidget):
    def new(self):
        self.gPodderPodcastDirectory.set_transient_for(self.parent_widget)
        if hasattr(self, 'custom_title'):
            self.main_window.set_title(self.custom_title)

        if not hasattr(self, 'add_podcast_list'):
            self.add_podcast_list = None

        self.providers_model = DirectoryProvidersModel(directory.PROVIDERS)
        self.podcasts_model = DirectoryPodcastsModel(self.on_can_subscribe_changed)
        self.current_provider = None

        self.setup_providers_treeview()
        self.setup_podcasts_treeview()

        accel = Gtk.AccelGroup()
        accel.connect(Gdk.KEY_Escape, 0, Gtk.AccelFlags.VISIBLE, self.on_escape)
        self.main_window.add_accel_group(accel)

        self._config.connect_gtk_window(self.main_window, 'podcastdirectory', True)

        self.btnBack.hide()
        self.main_window.show()

    def download_opml_file(self, filename):
        provider = directory.FixedOpmlFileProvider(filename)
        self.providers_model.add_provider(provider)
        self.tv_providers.set_cursor(len(self.providers_model) - 1)
        self.use_provider(provider, allow_back=False)

    def setup_podcasts_treeview(self):
        column = Gtk.TreeViewColumn('')
        cell = Gtk.CellRendererToggle()
        cell.set_fixed_size(48, -1)
        column.pack_start(cell, False)
        column.add_attribute(cell, 'active', DirectoryPodcastsModel.C_SELECTED)
        cell.connect('toggled', lambda cell, path: self.podcasts_model.toggle(path))
        self.tv_podcasts.append_column(column)

        column = Gtk.TreeViewColumn('')
        cell = Gtk.CellRendererText()
        cell.set_property('ellipsize', Pango.EllipsizeMode.END)
        column.pack_start(cell, True)
        column.add_attribute(cell, 'markup', DirectoryPodcastsModel.C_MARKUP)
        self.tv_podcasts.append_column(column)

        self.tv_podcasts.set_model(self.podcasts_model)
        self.podcasts_model.append((False, 'a', 'b', 'c', 'd'))

    def setup_providers_treeview(self):
        column = Gtk.TreeViewColumn('')
        cell = Gtk.CellRendererPixbuf()
        column.pack_start(cell, False)
        column.add_attribute(cell, 'pixbuf', DirectoryProvidersModel.C_ICON)
        cell = Gtk.CellRendererText()
        # cell.set_property('ellipsize', Pango.EllipsizeMode.END)
        column.pack_start(cell, True)
        column.add_attribute(cell, 'text', DirectoryProvidersModel.C_TEXT)
        column.add_attribute(cell, 'weight', DirectoryProvidersModel.C_WEIGHT)
        self.tv_providers.append_column(column)

        self.tv_providers.set_row_separator_func(self.providers_model.is_row_separator)

        self.tv_providers.set_model(self.providers_model)

        self.tv_providers.connect("row-activated", self.on_tv_providers_row_activated)

    def on_tv_providers_row_activated(self, treeview, path, column):
        it = self.providers_model.get_iter(path)

        for row in self.providers_model:
            row[DirectoryProvidersModel.C_WEIGHT] = Pango.Weight.NORMAL

        if it:
            self.providers_model.set_value(it, DirectoryProvidersModel.C_WEIGHT, Pango.Weight.BOLD)
            provider = self.providers_model.get_value(it, DirectoryProvidersModel.C_PROVIDER)
            self.use_provider(provider)

    def use_provider(self, provider, allow_back=True):
        self.podcasts_model.clear()
        self.current_provider = provider
        self.main_window.set_title(provider.name)

        if provider.kind == directory.Provider.PROVIDER_SEARCH:
            self.lb_search.set_text(_('Search:'))
            self.bt_search.set_label(_('Search'))
        elif provider.kind == directory.Provider.PROVIDER_URL:
            self.lb_search.set_text(_('URL:'))
            self.bt_search.set_label(_('Download'))
        elif provider.kind == directory.Provider.PROVIDER_FILE:
            self.lb_search.set_text(_('Filename:'))
            self.bt_search.set_label(_('Open'))
        elif provider.kind == directory.Provider.PROVIDER_STATIC:
            self.obtain_podcasts_with(provider.on_static)

        if provider.kind in (directory.Provider.PROVIDER_SEARCH,
                directory.Provider.PROVIDER_URL,
                directory.Provider.PROVIDER_FILE):
            self.en_query.set_text('')
            self.hb_text_entry.show()
            util.idle_add(self.en_query.grab_focus)
        else:
            self.hb_text_entry.hide()
        self.progressBar.set_fraction(0)
        self.progressLabel.set_label('')
        self.txtError.hide()
        self.stState.set_visible_child(self.stPodcasts)
        self.btnBack.set_visible(allow_back)

    def obtain_podcasts_with(self, callback):
        self.progressBar.set_fraction(0.1)
        self.progressLabel.set_text(_('Please wait while the podcast list is downloaded'))
        self.txtError.hide()
        self.stackProgressErrorPodcasts.set_visible_child(self.stProgError)
        self.selectbox.hide()
        self.btnOK.hide()

        original_provider = self.current_provider

        self.en_query.set_sensitive(False)
        self.bt_search.set_sensitive(False)
        self.podcasts_model.clear()

        @util.run_in_background
        def download_data():
            podcasts = []
            error = None
            try:
                podcasts = callback()
            except directory.JustAWarning as e:
                error = e
            except Exception as e:
                logger.warning(
                    'Got exception while loading podcasts: %r', e,
                    exc_info=True)
                error = e

            @util.idle_add
            def update_ui():
                self.progressBar.set_fraction(1)

                if original_provider == self.current_provider:
                    self.podcasts_model.load(podcasts or [])
                    if error:
                        self.progressLabel.set_text(_("Error"))
                        if isinstance(error, directory.JustAWarning):
                            self.txtError.get_buffer().set_text(error.warning)
                        else:
                            self.txtError.get_buffer().set_text(_("Error: %s\n\n%s") % (error, "".join(traceback.format_exception(error))))
                        self.txtError.show()
                    elif not podcasts:
                        self.progressLabel.set_text(_("No results"))
                        self.txtError.get_buffer().set_text(_("Sorry, no podcasts were found"))
                        self.txtError.show()
                    else:
                        self.stackProgressErrorPodcasts.set_visible_child(self.sw_podcasts)
                        self.selectbox.show()
                        self.btnOK.show()
                else:
                    logger.warning('Ignoring update from old thread')

                self.en_query.set_sensitive(True)
                self.bt_search.set_sensitive(True)
                if self.en_query.get_realized():
                    self.en_query.grab_focus()

    def on_bt_search_clicked(self, widget):
        if self.current_provider is None:
            return
        query = self.en_query.get_text()

        @self.obtain_podcasts_with
        def load_data():
            if self.current_provider.kind == directory.Provider.PROVIDER_SEARCH:
                return self.current_provider.on_search(query)
            elif self.current_provider.kind == directory.Provider.PROVIDER_URL:
                return self.current_provider.on_url(query)
            elif self.current_provider.kind == directory.Provider.PROVIDER_FILE:
                return self.current_provider.on_file(query)

    def on_can_subscribe_changed(self, can_subscribe):
        self.btnOK.set_sensitive(can_subscribe)

    def on_btnSelectAll_clicked(self, widget, *args):
        self.podcasts_model.set_selection_to(True)

    def on_btnSelectNone_clicked(self, widget, *args):
        self.podcasts_model.set_selection_to(False)

    def on_btnOK_clicked(self, widget, *args):
        urls = self.podcasts_model.get_selected_podcasts()

        self.main_window.destroy()

        # add channels that have been selected
        if self.add_podcast_list is not None:
            self.add_podcast_list(urls)

    def on_btnCancel_clicked(self, widget, *args):
        self.main_window.destroy()

    def on_btnBack_clicked(self, widget, *args):
        self.stState.set_visible_child(self.stProviders)
        widget.hide()
        self.selectbox.hide()
        self.btnOK.hide()

    def on_escape(self, *args, **kwargs):
        if self.stState.get_visible_child() == self.stProviders:
            self.main_window.destroy()
        else:
            self.on_btnBack_clicked(self.btnBack)



# -*- coding: utf-8 -*-
#
# gPodder - A media aggregator and podcast client
# Copyright (c) 2005-2018 The gPodder Team
#
# gPodder is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# gPodder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# Windows 7 taskbar progress
# Sean Munkel; 2013-01-05

import ctypes
import functools
import logging
from ctypes import (HRESULT, POINTER, Structure, alignment, c_int, c_uint,
                    c_ulong, c_ulonglong, c_ushort, c_wchar_p, sizeof)
from ctypes.wintypes import tagRECT

from comtypes import COMMETHOD, GUID, IUnknown, client, wireHWND

import gpodder

import gi  # isort:skip
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk  # isort:skip


_ = gpodder.gettext


__title__ = _('Show download progress on the taskbar')
__description__ = _('Displays the progress on the Windows taskbar.')
__authors__ = 'Sean Munkel <seanmunkel@gmail.com>'
__category__ = 'desktop-integration'
__only_for__ = 'win32'

logger = logging.getLogger(__name__)

WSTRING = c_wchar_p
# values for enumeration 'TBPFLAG'
TBPF_NOPROGRESS = 0
TBPF_INDETERMINATE = 1
TBPF_NORMAL = 2
TBPF_ERROR = 4
TBPF_PAUSED = 8
TBPFLAG = c_int  # enum
# values for enumeration 'TBATFLAG'
TBATF_USEMDITHUMBNAIL = 1
TBATF_USEMDILIVEPREVIEW = 2
TBATFLAG = c_int  # enum
# return code
S_OK = HRESULT(0).value


class tagTHUMBBUTTON(Structure):
    _fields_ = [
        ('dwMask', c_ulong),
        ('iId', c_uint),
        ('iBitmap', c_uint),
        ('hIcon', POINTER(IUnknown)),
        ('szTip', c_ushort * 260),
        ('dwFlags', c_ulong)]


class ITaskbarList(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID('{56FDF342-FD6D-11D0-958A-006097C9A090}')
    _idlflags_ = []
    _methods_ = [
        COMMETHOD([], HRESULT, 'HrInit'),
        COMMETHOD([], HRESULT, 'AddTab',
                  (['in'], c_int, 'hwnd')),
        COMMETHOD([], HRESULT, 'DeleteTab',
                  (['in'], c_int, 'hwnd')),
        COMMETHOD([], HRESULT, 'ActivateTab',
                  (['in'], c_int, 'hwnd')),
        COMMETHOD([], HRESULT, 'SetActivateAlt',
                  (['in'], c_int, 'hwnd'))]


class ITaskbarList2(ITaskbarList):
    _case_insensitive_ = True
    _iid_ = GUID('{602D4995-B13A-429B-A66E-1935E44F4317}')
    _idlflags_ = []
    _methods_ = [
        COMMETHOD([], HRESULT, 'MarkFullscreenWindow',
                  (['in'], c_int, 'hwnd'),
                  (['in'], c_int, 'fFullscreen'))]


class ITaskbarList3(ITaskbarList2):
    _case_insensitive_ = True
    _iid_ = GUID('{EA1AFB91-9E28-4B86-90E9-9E9F8A5EEFAF}')
    _idlflags_ = []
    _methods_ = [
        COMMETHOD([], HRESULT, 'SetProgressValue',
                  (['in'], c_int, 'hwnd'),
                  (['in'], c_ulonglong, 'ullCompleted'),
                  (['in'], c_ulonglong, 'ullTotal')),
        COMMETHOD([], HRESULT, 'SetProgressState',
                  (['in'], c_int, 'hwnd'),
                  (['in'], TBPFLAG, 'tbpFlags')),
        COMMETHOD([], HRESULT, 'RegisterTab',
                  (['in'], c_int, 'hwndTab'),
                  (['in'], wireHWND, 'hwndMDI')),
        COMMETHOD([], HRESULT, 'UnregisterTab',
                  (['in'], c_int, 'hwndTab')),
        COMMETHOD([], HRESULT, 'SetTabOrder',
                  (['in'], c_int, 'hwndTab'),
                  (['in'], c_int, 'hwndInsertBefore')),
        COMMETHOD([], HRESULT, 'SetTabActive',
                  (['in'], c_int, 'hwndTab'),
                  (['in'], c_int, 'hwndMDI'),
                  (['in'], TBATFLAG, 'tbatFlags')),
        COMMETHOD([], HRESULT, 'ThumbBarAddButtons',
                  (['in'], c_int, 'hwnd'),
                  (['in'], c_uint, 'cButtons'),
                  (['in'], POINTER(tagTHUMBBUTTON), 'pButton')),
        COMMETHOD([], HRESULT, 'ThumbBarUpdateButtons',
                  (['in'], c_int, 'hwnd'),
                  (['in'], c_uint, 'cButtons'),
                  (['in'], POINTER(tagTHUMBBUTTON), 'pButton')),
        COMMETHOD([], HRESULT, 'ThumbBarSetImageList',
                  (['in'], c_int, 'hwnd'),
                  (['in'], POINTER(IUnknown), 'himl')),
        COMMETHOD([], HRESULT, 'SetOverlayIcon',
                  (['in'], c_int, 'hwnd'),
                  (['in'], POINTER(IUnknown), 'hIcon'),
                  (['in'], WSTRING, 'pszDescription')),
        COMMETHOD([], HRESULT, 'SetThumbnailTooltip',
                  (['in'], c_int, 'hwnd'),
                  (['in'], WSTRING, 'pszTip')),
        COMMETHOD([], HRESULT, 'SetThumbnailClip',
                  (['in'], c_int, 'hwnd'),
                  (['in'], POINTER(tagRECT), 'prcClip'))]


assert sizeof(tagTHUMBBUTTON) in [540, 552], sizeof(tagTHUMBBUTTON)
assert alignment(tagTHUMBBUTTON) in [4, 8], alignment(tagTHUMBBUTTON)


def consume_events():
    """Consume pending events."""
    while Gtk.events_pending():
        Gtk.main_iteration()


# based on http://stackoverflow.com/a/1744503/905256
class gPodderExtension:
    def __init__(self, container):
        self.container = container
        self.window_handle = None
        self.restart_warning = True

    def on_load(self):
        self.taskbar = client.CreateObject(
            '{56FDF344-FD6D-11d0-958A-006097C9A090}',
            interface=ITaskbarList3)
        ret = self.taskbar.HrInit()
        if ret != S_OK:
            logger.warning("taskbar.HrInit failed: %r", ret)
            del self.taskbar

    def on_unload(self):
        # let the window change state? otherwise gpodder is stuck on exit
        # (tested on windows 7 pro)
        consume_events()
        if self.taskbar is not None:
            self.taskbar.SetProgressState(self.window_handle, TBPF_NOPROGRESS)
            # let the taskbar change state otherwise gpodder is stuck on exit
            # (tested on windows 7 pro)
            consume_events()

    def on_ui_object_available(self, name, ui_object):
        def callback(self, window, *args):
            ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
            ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object]
            win_gpointer = ctypes.pythonapi.PyCapsule_GetPointer(window.get_window().__gpointer__, None)
            gdkdll = ctypes.CDLL("libgdk-3-0.dll")
            self.window_handle = gdkdll.gdk_win32_window_get_handle(win_gpointer)
            ret = self.taskbar.ActivateTab(self.window_handle)
            if ret != S_OK:
                logger.warning("taskbar.ActivateTab failed: %r", ret)
                del self.taskbar

        if name == 'gpodder-gtk':
            ui_object.main_window.connect('realize',
                    functools.partial(callback, self))

    def on_download_progress(self, progress):
        if not self.taskbar:
            return
        if self.window_handle is None:
            if not self.restart_warning:
                return
            logger.warning("No window handle available, a restart max fix this")
            self.restart_warning = False
            return
        if 0 < progress < 1:
            self.taskbar.SetProgressState(self.window_handle, TBPF_NORMAL)
            self.taskbar.SetProgressValue(self.window_handle,
                    int(progress * 100), 100)
        else:
            self.taskbar.SetProgressState(self.window_handle, TBPF_NOPROGRESS)



# -*- coding: utf-8 -*-
# Searches podverse (podverse.fm) database for podcasts

# (c) 2024 Eric Le Lay <elelay.fr:contact>
# Released under the same license terms as gPodder itself.

# Inspired by gpodder-core plugin "podverse", by kirbylife <hola@kirbylife.dev>
# https://github.com/gpodder/gpodder-core/blob/master/src/gpodder/plugins/podverse.py

import logging
from urllib.parse import quote_plus

import requests

import gpodder
from gpodder.directory import PROVIDERS, DirectoryEntry, JustAWarning, Provider
from gpodder.util import urlopen

logger = logging.getLogger(__name__)

_ = gpodder.gettext

__title__ = _('Search Podverse')
__description__ = _('Search podverse podcast index')
__authors__ = 'Eric Le Lay <elelay.fr:contact>'
__doc__ = 'https://gpodder.github.io/docs/extensions/podverse.html'


class PodverseDirectoryProvider(Provider):
    def __init__(self):
        self.name = _('Podverse search')
        self.kind = Provider.PROVIDER_SEARCH
        self.icon = 'directory-podverse.png'

    def on_search(self, query):
        if len(query) < 3:
            raise JustAWarning(_("Please provide at least 3 characters"))
        # see https://api.podverse.fm/api/v1/swagger#operations-podcast-getPodcasts
        json_url = f"https://api.podverse.fm/api/v1/podcast?page=1&searchTitle={quote_plus(query)}&sort=top-past-week"
        response = urlopen(json_url, headers={"accept": "application/json"})
        json_data = response.json()
        if response.status_code != requests.codes.ok:
            raise Exception(_("Error searching: %s") % json_data.get("message"))

        # contrary to swagger api we get a [results_page, total_length] 2 element list
        # See code in https://github.com/podverse/podverse-api/blob/develop/src/controllers/podcast.ts#L311
        if isinstance(json_data, list) and len(json_data) == 2 and isinstance(json_data[0], list):
            logger.debug("Search for %s yields %i results, of which we display %i",
                query, json_data[1], len(json_data[0]))
            json_data = json_data[0]
        else:
            logger.debug("Search for %s yields %i results", query, len(json_data))

        return [
            DirectoryEntry(e["title"],
                           e["feedUrls"][0]["url"],
                           image=e["imageUrl"],
                           description=e["description"])
            for e in json_data if not e["credentialsRequired"]
        ]


class gPodderExtension:
    """Register and unregister a podverse search provider."""

    def __init__(self, container):
        pass

    def on_load(self):
        logger.info('Registering Podverse.')
        PROVIDERS.append(None)
        PROVIDERS.append(PodverseDirectoryProvider)

    def on_unload(self):
        logger.info('Unregistering Podverse.')
        try:
            PROVIDERS.remove(PodverseDirectoryProvider)
        except Exception:
            logger.exception("Unable to remove PodverseDirectoryProvider")



# -*- coding: utf-8 -*-
#
# gPodder - A media aggregator and podcast client
# Copyright (c) 2005-2018 The gPodder Team
#
# gPodder is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# gPodder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import time

from gi.repository import GLib, Gtk, Pango

import gpodder
from gpodder import util
from gpodder.gtkui.widgets import SpinningProgressIndicator

_ = gpodder.gettext


class ProgressIndicator(object):
    # Delayed time until window is shown (for short operations)
    DELAY = 500

    # Time between GUI updates after window creation
    INTERVAL = 250

    def __init__(self, title, subtitle=None, cancellable=False, parent=None, max_ticks=None):
        self.title = title
        self.subtitle = subtitle
        self.cancellable = True if cancellable else False
        self.cancel_callback = cancellable
        self.cancel_id = 0
        self.cancelled = False
        self.next_update = time.time() + (self.DELAY / 1000)
        self.parent = parent
        self.dialog = None
        self.progressbar = None
        self.indicator = None
        self._initial_message = None
        self._initial_progress = None
        self._progress_set = False
        # use timeout_add, not util.idle_timeout_add, so it updates before Gtk+ redraws the dialog
        self.source_id = GLib.timeout_add(self.DELAY, self._create_progress)

        self.set_max_ticks(max_ticks)

    def set_max_ticks(self, max_ticks):
        self.max_ticks = max_ticks
        self.tick_counter = 0
        if max_ticks is not None:
            self.on_message('0 / %d' % max_ticks)

    def _on_delete_event(self, window, event):
        if self.cancellable:
            self.dialog.response(Gtk.ResponseType.CANCEL)
            self.cancellable = False
            self.cancelled = True
        return True

    def _create_progress(self):
        self.dialog = Gtk.MessageDialog(self.parent,
                0, 0, Gtk.ButtonsType.CANCEL, self.subtitle or self.title)
        self.dialog.set_modal(True)
        self.dialog.connect('delete-event', self._on_delete_event)
        if self.cancellable:
            def cancel_callback(dialog, response):
                self.cancellable = False
                self.cancelled = True
                self.dialog.set_deletable(False)
                self.dialog.set_response_sensitive(Gtk.ResponseType.CANCEL, False)
                if callable(self.cancel_callback):
                    self.cancel_callback(dialog, response)
            self.cancel_id = self.dialog.connect('response', cancel_callback)
        self.dialog.set_title(self.title)
        self.dialog.set_deletable(self.cancellable)

        # Avoid selectable text (requires PyGTK >= 2.22)
        if hasattr(self.dialog, 'get_message_area'):
            for label in self.dialog.get_message_area():
                if isinstance(label, Gtk.Label):
                    label.set_selectable(False)

        self.dialog.set_response_sensitive(Gtk.ResponseType.CANCEL, self.cancellable)

        self.progressbar = Gtk.ProgressBar()
        self.progressbar.set_show_text(True)
        self.progressbar.set_ellipsize(Pango.EllipsizeMode.END)

        # If the window is shown after the first update, set the progress
        # info so that when the window appears, data is there already
        if self._initial_progress is not None:
            self.progressbar.set_fraction(self._initial_progress)
        if self._initial_message is not None:
            self.progressbar.set_text(self._initial_message)

        self.dialog.vbox.add(self.progressbar)
        self.indicator = SpinningProgressIndicator()
        self.dialog.set_image(self.indicator)
        self.dialog.show_all()

        self._update_gui()

        # previous self.source_id timeout is removed when this returns False
        # use timeout_add, not util.idle_timeout_add, so it updates before Gtk+ redraws the dialog
        self.source_id = GLib.timeout_add(self.INTERVAL, self._update_gui)
        return False

    def _update_gui(self):
        if self.indicator:
            self.indicator.step_animation()
        if not self._progress_set and self.progressbar:
            self.progressbar.pulse()
        self.next_update = time.time() + (self.INTERVAL / 1000)
        return True

    def on_message(self, message):
        if self.progressbar:
            self.progressbar.set_text(message)
        else:
            self._initial_message = message

    def on_progress(self, progress):
        self._progress_set = True
        if self.progressbar:
            self.progressbar.set_fraction(progress)
        else:
            self._initial_progress = progress

    def on_tick(self, final=False):
        if final:
            # Dialog is no longer cancellable
            self.cancellable = False
            if self.dialog is not None:
                self.dialog.set_response_sensitive(Gtk.ResponseType.CANCEL, False)
                self.dialog.set_deletable(False)
            elif 2 * (time.time() - (self.next_update - (self.DELAY / 1000))) > (self.DELAY / 1000):
                # Assume final operation will take as long as all ticks and open dialog
                if self.source_id:
                    GLib.source_remove(self.source_id)
                self._create_progress()

        if self.max_ticks is not None and not final:
            self.tick_counter += 1

        if time.time() >= self.next_update or (final and self.dialog):
            if isinstance(final, str):
                self.on_message(final)
                self.on_progress(1.0)
            elif self.max_ticks is not None:
                self.on_message('%d / %d' % (self.tick_counter, self.max_ticks))
                self.on_progress(self.tick_counter / self.max_ticks)

            # Allow UI to redraw.
            util.idle_add(Gtk.main_quit)
            # self._create_progress() or self._update_gui() is called by a timer to update the dialog
            Gtk.main()

            if self.cancelled:
                return False
        return True

    def on_finished(self):
        if self.dialog is not None:
            if self.cancel_id:
                self.dialog.disconnect(self.cancel_id)
            self.dialog.destroy()
        if self.source_id:
            GLib.source_remove(self.source_id)



# -*- coding: utf-8 -*-
# Show publishing statistics for subscriptions.
# Released under the same license terms as gPodder itself.

import locale
import time
from time import localtime, strftime

import gpodder
from gpodder import config

import gi  # isort:skip
from gi.repository import Gtk  # isort:skip
gi.require_version('Gtk', '3.0')  # isort:skip

_ = gpodder.gettext

__title__ = _('Subscription Statistics')
__description__ = _('Show publishing statistics for subscriptions.')
__only_for__ = 'gtk'
__doc__ = 'https://gpodder.github.io/docs/extensions/subscription_stats.html'
__authors__ = 'Brand Huntsman <http://qzx.com/mail/> and \
               Nuno Dias <Nuno.Dias+gpodder@gmail.com>'


class gPodderExtension:
    def __init__(self, container):
        self.container = container

    def on_ui_object_available(self, name, ui_object):
        if name == 'gpodder-gtk':
            self.gpodder = ui_object

    def on_create_menu(self):
        # extras menu
        return [(_("Subscription Statistics"), self.open_dialog)]

    def add_page(self, notebook, category, channels):
        scrolled = Gtk.ScrolledWindow()

        store = Gtk.ListStore(str, float, str, str, int)
        for average, name, edate, paused in channels:
            last = strftime('%x', localtime(edate))
            store.append([
                locale.format_string('%.1f', round(average, 1)) if average > 0 else '?',
                average, ('❚❚ ' if paused else '') + name, last, edate,
            ])

        tree = Gtk.TreeView(model=store)
        scrolled.add(tree)

        lastcell = Gtk.CellRendererText()
        lastcolumn = Gtk.TreeViewColumn(_('Updated'))
        lastcolumn.set_sort_column_id(4)
        lastcolumn.pack_end(lastcell, True)
        lastcolumn.add_attribute(lastcell, 'text', 3)
        tree.append_column(lastcolumn)

        dayscell = Gtk.CellRendererText()
        dayscell.set_property('xalign', 1)
        dayscolumn = Gtk.TreeViewColumn(_('Days'))
        dayscolumn.set_sort_column_id(1)
        dayscolumn.pack_start(dayscell, True)
        dayscolumn.add_attribute(dayscell, 'text', 0)
        tree.append_column(dayscolumn)

        channelcell = Gtk.CellRendererText()
        channelcolumn = Gtk.TreeViewColumn(_('Podcast'))
        channelcolumn.set_sort_column_id(2)
        channelcolumn.pack_start(channelcell, True)
        channelcolumn.add_attribute(channelcell, 'text', 2)
        channelcolumn.set_expand(True)
        tree.append_column(channelcolumn)

        notebook.append_page(scrolled,
                             Gtk.Label('%d %s' % (len(channels), category)))

    def open_dialog(self):
        db = self.gpodder.db

        # get all channels
        channels = []
        with db.lock:
            cur = db.cursor()
            cur.execute(
                        'SELECT id, title, pause_subscription FROM %s'
                        % db.TABLE_PODCAST)
            while True:
                row = cur.fetchone()
                if row is None:
                    break
                channels.append(row)
            cur.close()

        # get average time between episodes per channel
        now = int(time.time())
        nr_paused = 0
        daily = []
        weekly = []
        monthly = []
        yearly = []
        for channel_id, channel_name, paused in channels:
            if paused:
                nr_paused += 1

            total = 0
            nr_episodes = 0
            prev = now
            with db.lock:
                cur = db.cursor()
                cur.execute('SELECT published FROM %s WHERE podcast_id = %d \
                            ORDER BY published DESC LIMIT 25'
                            % (db.TABLE_EPISODE, channel_id))
                while True:
                    row = cur.fetchone()
                    if row is None:
                        break
                    if total == 0:
                        edate = row[0]
                    total += (prev - row[0])
                    nr_episodes += 1
                    prev = row[0]
                cur.close()

            average = (total / nr_episodes) / (24 * 60 * 60) \
                if nr_episodes > 0 else 0

            if average == 0:
                yearly.append([average, channel_name, edate, paused])
            elif average <= 2:
                daily.append([average, channel_name, edate, paused])
            elif average <= 14:
                weekly.append([average, channel_name, edate, paused])
            elif average <= 61:
                monthly.append([average, channel_name, edate, paused])
            else:
                yearly.append([average, channel_name, edate, paused])

        # sort by averages
        daily.sort(key=lambda e: e[0])
        weekly.sort(key=lambda e: e[0])
        monthly.sort(key=lambda e: e[0])
        yearly.sort(key=lambda e: e[0])

        # open dialog
        dlg = Gtk.Dialog(_('Subscription Statistics'),
                         self.gpodder.main_window)
        dlg.set_default_size(420, 600)
        dlg.set_resizable(True)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.set_border_width(0)

        label = Gtk.Label(_('%d subscriptions (%d paused)') % (len(channels),
                          nr_paused))
        label.set_padding(0, 10)
        box.add(label)

        notebook = Gtk.Notebook()
        notebook.set_vexpand(True)
        notebook.set_scrollable(True)
        notebook.set_show_border(False)
        self.add_page(notebook, _('daily'), daily)
        self.add_page(notebook, _('weekly'), weekly)
        self.add_page(notebook, _('monthly'), monthly)
        self.add_page(notebook, _('yearly'), yearly)
        box.add(notebook)

        conf = config.Config(gpodder.config_file)
        label = Gtk.Label(_('Average days between the last %d episodes.') %
                          (conf.limit.episodes if conf.limit.episodes < 25
                           else 25))
        label.set_line_wrap(True)
        label.set_padding(0, 5)
        box.add(label)

        button = dlg.add_button(_('_Close'), Gtk.ResponseType.OK)
        button.set_margin_right(5)
        button.set_margin_bottom(5)
        dlg.connect("response", lambda w, r: dlg.destroy())

        dlg.vbox.pack_start(box, True, True, 0)
        dlg.vbox.show_all()
        dlg.show()



# -*- coding: utf-8 -*-
#
# gPodder - A media aggregator and podcast client
# Copyright (c) 2005-2011 Thomas Perl and the gPodder Team
#
# gPodder is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# gPodder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os

import gpodder
from gpodder import util
from gpodder.sync import (episode_filename_on_device,
                          episode_foldername_on_device)

import gi  # isort:skip
gi.require_version('Gio', '2.0')  # isort:skip
from gi.repository import Gio, GLib  # isort:skip

_ = gpodder.gettext


logger = logging.getLogger(__name__)


class gPodderDevicePlaylist(object):
    def __init__(self, config, playlist_name):
        self._config = config
        self.linebreak = '\r\n'
        self.playlist_file = (
            util.sanitize_filename(playlist_name, self._config.device_sync.max_filename_length)
            + '.' + self._config.device_sync.playlists.extension)
        device_folder = util.new_gio_file(self._config.device_sync.device_folder)
        self.playlist_folder = device_folder.resolve_relative_path(self._config.device_sync.playlists.folder)

        self.mountpoint = None
        try:
            self.mountpoint = self.playlist_folder.find_enclosing_mount().get_root()
        except GLib.Error as err:
            logger.error('find_enclosing_mount folder %s failed: %s', self.playlist_folder.get_uri(), err.message)

        if not self.mountpoint:
            self.mountpoint = self.playlist_folder
            logger.warning('could not find mount point for MP3 player - using %s as MP3 player root', self.mountpoint.get_uri())
        self.playlist_absolute_filename = self.playlist_folder.resolve_relative_path(self.playlist_file)

    def build_extinf(self, filename):
        # TODO: Windows playlists
        #        if self._config.mp3_player_playlist_win_path:
        #            filename = filename.replace('\\', os.sep)

        #        # rebuild the whole filename including the mountpoint
        #        if self._config.device_sync.playlist_absolute_path:
        #            absfile = os.path.join(self.mountpoint,filename)
        #        else: #TODO: Test rel filenames
        #            absfile = util.rel2abs(filename, os.path.dirname(self.playlist_file))

        # fallback: use the basename of the file
        (title, extension) = os.path.splitext(os.path.basename(filename))

        return "#EXTINF:0,%s%s" % (title.strip(), self.linebreak)

    def read_m3u(self):
        """Read all files from the existing playlist."""
        tracks = []
        logger.info("Read data from the playlistfile %s" % self.playlist_absolute_filename.get_uri())
        if self.playlist_absolute_filename.query_exists():
            stream = Gio.DataInputStream.new(self.playlist_absolute_filename.read())
            while True:
                line = stream.read_line_utf8()[0]
                if not line:
                    break
                if not line.startswith('#EXT'):
                    tracks.append(line.rstrip('\r\n'))
            stream.close()
        return tracks

    def get_filename_for_playlist(self, episode):
        """Get the filename for the given episode for the playlist."""
        return episode_filename_on_device(self._config, episode)

    def get_absolute_filename_for_playlist(self, episode):
        """Get the filename including full path for the given episode for the playlist."""
        filename = self.get_filename_for_playlist(episode)
        foldername = episode_foldername_on_device(self._config, episode)
        if foldername:
            filename = os.path.join(foldername, filename)
        if self._config.device_sync.playlists.use_absolute_path:
            filename = os.path.join(util.relpath(self._config.device_sync.device_folder, self.mountpoint.get_uri()), filename)
        return filename

    def write_m3u(self, episodes):
        """Write the list into the playlist on the device."""
        logger.info('Writing playlist file: %s', self.playlist_file)
        if not util.make_directory(self.playlist_folder):
            raise IOError(_('Folder %s could not be created.') % self.playlist_folder, _('Error writing playlist'))
        else:
            # work around libmtp devices potentially having limited capabilities for partial writes
            is_mtp = self.playlist_folder.get_uri().startswith("mtp://")
            tempfile = None
            if is_mtp:
                tempfile = Gio.File.new_tmp()
                fs = tempfile[1].get_output_stream()
            else:
                fs = self.playlist_absolute_filename.replace(None, False, Gio.FileCreateFlags.NONE)

            os = Gio.DataOutputStream.new(fs)
            os.put_string('#EXTM3U%s' % self.linebreak)
            for current_episode in episodes:
                filename = self.get_filename_for_playlist(current_episode)
                os.put_string(self.build_extinf(filename))
                filename = self.get_absolute_filename_for_playlist(current_episode)
                os.put_string(filename)
                os.put_string(self.linebreak)
            os.close()

            if is_mtp:
                try:
                    tempfile[0].copy(self.playlist_absolute_filename, Gio.FileCopyFlags.OVERWRITE)
                except GLib.Error as err:
                    logger.error('copying playlist to mtp device file %s failed: %s',
                        self.playlist_absolute_filename.get_uri(), err.message)
                tempfile[0].delete()



# -*- coding: utf-8 -*-
#
# gPodder - A media aggregator and podcast client
# Copyright (c) 2005-2018 The gPodder Team
#
# gPodder is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# gPodder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gdk, Gio, Gtk

import gpodder
from gpodder import util
from gpodder.gtkui.interface.common import BuilderWidget

_ = gpodder.gettext


class gPodderChannel(BuilderWidget):
    MAX_SIZE = 120

    def new(self):
        self.show_on_cover_load = True

        self.gPodderChannel.set_transient_for(self.parent_widget)
        self.title_label.set_text(self.channel.title)
        self.labelURL.set_text(self.channel.url)
        self.skip_feed_update_switch.set_active(self.channel.pause_subscription)
        self.enable_device_sync_switch.set_active(self.channel.sync_to_mp3_player)

        self.section_list = Gtk.ListStore(str)
        active_index = 0
        for index, section in enumerate(sorted(self.sections)):
            self.section_list.append([section])
            if section == self.channel.section:
                active_index = index
        self.combo_section.set_model(self.section_list)
        cell_renderer = Gtk.CellRendererText()
        self.combo_section.pack_start(cell_renderer, True)
        self.combo_section.add_attribute(cell_renderer, 'text', 0)
        self.combo_section.set_active(active_index)

        self.strategy_list = Gtk.ListStore(str, int)
        active_index = 0
        for index, (checked, strategy_id, strategy) in \
                enumerate(self.channel.get_download_strategies()):
            self.strategy_list.append([strategy, strategy_id])
            if checked:
                active_index = index
        self.combo_strategy.set_model(self.strategy_list)
        cell_renderer = Gtk.CellRendererText()
        self.combo_strategy.pack_start(cell_renderer, True)
        self.combo_strategy.add_attribute(cell_renderer, 'text', 0)
        self.combo_strategy.set_active(active_index)

        self.LabelDownloadTo.set_text(self.channel.save_dir)
        self.website_label.set_markup('<a href="{}">{}</a>'.format(
            self.channel.link, self.channel.link)
            if self.channel.link else '')
        self.website_label.connect('activate-link', lambda label, url: util.open_website(url))

        if self.channel.auth_username:
            self.FeedUsername.set_text(self.channel.auth_username)
        if self.channel.auth_password:
            self.FeedPassword.set_text(self.channel.auth_password)

        # Cover image
        ag = Gio.SimpleActionGroup()
        open_cover_action = Gio.SimpleAction.new("openCover", None)
        open_cover_action.connect('activate', self.on_open_cover_activate)
        ag.add_action(open_cover_action)
        refresh_cover_action = Gio.SimpleAction.new("refreshCover", None)
        refresh_cover_action.connect('activate', self.on_refresh_cover_activate)
        ag.add_action(refresh_cover_action)
        self.main_window.insert_action_group("channel", ag)

        cover_menu = Gio.Menu()
        cover_menu.append("Change cover image", "channel.openCover")
        cover_menu.append("Refresh image", "channel.refreshCover")

        self.cover_menubutton.set_menu_model(cover_menu)

        self.cover_downloader.register('cover-available', self.cover_download_finished)
        self.cover_downloader.request_cover(self.channel)

        if self.channel._update_error:
            err = '\n\n' + (_('ERROR: %s') % self.channel._update_error)
        else:
            err = ''
        self.channel_description.set_text(util.remove_html_tags(self.channel.description) + err)

        # Add Drag and Drop Support
        flags = Gtk.DestDefaults.ALL
        targets = [Gtk.TargetEntry.new('text/uri-list', 0, 2), Gtk.TargetEntry.new('text/plain', 0, 4)]
        actions = Gdk.DragAction.DEFAULT | Gdk.DragAction.COPY
        self.imgCover.drag_dest_set(flags, targets, actions)
        self.imgCover.connect('drag_data_received', self.drag_data_received)
        border = 6
        size = self.MAX_SIZE + border * 2
        self.imgCover.set_size_request(size, size)

        # Title save button state
        self.title_save_button_saves = True

        self._config.connect_gtk_window(self.gPodderChannel, 'channel_editor', True)

        gpodder.user_extensions.on_ui_object_available('channel-gtk', self)

        result = gpodder.user_extensions.on_channel_settings(self.channel)
        if result:
            for label, callback in result:
                sw = Gtk.ScrolledWindow()
                sw.add(callback(self.channel))
                sw.show_all()
                self.notebookChannelEditor.append_page(sw, Gtk.Label(label))

    def on_button_add_section_clicked(self, widget):
        text = self.show_text_edit_dialog(_('Add section'), _('New section:'),
            affirmative_text=_('_Add'))

        if text is not None:
            for index, (section,) in enumerate(self.section_list):
                if text == section:
                    self.combo_section.set_active(index)
                    return

            self.section_list.append([text])
            self.combo_section.set_active(len(self.section_list) - 1)

    def on_open_cover_activate(self, action, *args):
        dlg = Gtk.FileChooserDialog(
            title=_('Select new podcast cover artwork'),
            parent=self.gPodderChannel,
            action=Gtk.FileChooserAction.OPEN)
        dlg.add_button(_('_Cancel'), Gtk.ResponseType.CANCEL)
        dlg.add_button(_('_Open'), Gtk.ResponseType.OK)

        if dlg.run() == Gtk.ResponseType.OK:
            url = dlg.get_uri()
            self.clear_cover_cache(self.channel.url)
            self.cover_downloader.replace_cover(self.channel, custom_url=url)

        dlg.destroy()

    def on_refresh_cover_activate(self, action, *args):
        self.clear_cover_cache(self.channel.url)
        self.cover_downloader.replace_cover(self.channel, custom_url=False)

    def cover_download_finished(self, channel, pixbuf):
        def set_cover(channel, pixbuf):
            if self.channel == channel:
                if pixbuf is not None:
                    self.imgCover.set_from_pixbuf(util.scale_pixbuf(pixbuf, self.MAX_SIZE))
                if self.show_on_cover_load:
                    self.main_window.show()
                    self.show_on_cover_load = False

        util.idle_add(set_cover, channel, pixbuf)

    def drag_data_received(self, widget, content, x, y, sel, ttype, time):
        files = sel.get_text().strip().split('\n')
        if len(files) != 1:
            self.show_message(
                _('You can only drop a single image or URL here.'),
                _('Drag and drop'))
            return

        file = files[0]

        if file.startswith('file://') or file.startswith('http://') or file.startswith('https://'):
            self.clear_cover_cache(self.channel.url)
            self.cover_downloader.replace_cover(self.channel, custom_url=file)
            return

        self.show_message(
            _('You can only drop local files and http:// URLs here.'),
            _('Drag and drop'))

    def on_gPodderChannel_destroy(self, widget, *args):
        self.cover_downloader.unregister('cover-available', self.cover_download_finished)

    # Title editing callbacks
    def on_title_edit_button_clicked(self, button):
        self.title_save_button_saves = True
        self.title_save_button.set_label(_("_Save"))
        self.title_stack.set_visible_child(self.title_edit_box)
        self.title_entry.set_text(self.title_label.get_text())
        self.title_entry.grab_focus()

    def on_title_entry_changed(self, entry):
        if len(entry.get_text()) > 0:
            self.title_save_button_saves = True
            self.title_save_button.set_label(_("_Save"))
        else:
            self.title_save_button_saves = False
            self.title_save_button.set_label(_("Cancel"))

    def on_title_entry_icon_press(self, entry, icon_pos, *args):
        self.title_entry.set_text("")

    def on_title_save_button_clicked(self, button):
        if self.title_save_button_saves:
            self.title_label.set_text(self.title_entry.get_text())
        self.title_stack.set_visible_child(self.title_box)

    def on_feed_url_copy_button_clicked(self, button):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(self.channel.url, -1)

    def on_open_folder_button_clicked(self, button):
        util.gui_open(self.channel.save_dir, gui=self)

    def on_row_activated(self, listbox, row, *args):
        # Find the correct widget in the row to activate
        def _do(w, *args):
            if w.get_name().startswith("no_activation"):
                return
            elif isinstance(w, Gtk.Box):
                w.foreach(_do)
            elif isinstance(w, Gtk.ComboBox):
                w.popup()
            elif isinstance(w, Gtk.Entry):
                w.grab_focus()
            elif isinstance(w, Gtk.Switch):
                w.set_state(not w.get_state())
            elif isinstance(w, Gtk.Button):
                w.emit("clicked")
        row.foreach(_do)

    def on_btnCancel_clicked(self, widget, *args):
        self.main_window.destroy()

    def on_btnOK_clicked(self, widget, *args):
        self.channel.pause_subscription = self.skip_feed_update_switch.get_state()
        self.channel.sync_to_mp3_player = self.enable_device_sync_switch.get_state()
        self.channel.rename(self.title_label.get_text())
        self.channel.auth_username = self.FeedUsername.get_text().strip()
        self.channel.auth_password = self.FeedPassword.get_text()

        self.cover_downloader.unregister('cover-available', self.cover_download_finished)
        self.clear_cover_cache(self.channel.url)
        self.cover_downloader.request_cover(self.channel)

        new_section = self.section_list[self.combo_section.get_active()][0]
        if self.channel.section != new_section:
            self.channel.section = new_section
            section_changed = True
        else:
            section_changed = False

        new_strategy = self.strategy_list[self.combo_strategy.get_active()][1]
        self.channel.set_download_strategy(new_strategy)

        self.channel.save()

        self.main_window.destroy()

        self.update_podcast_list_model(selected=True,
                sections_changed=section_changed)



#
# -*- coding: utf-8 -*-
#
# gPodder - A media aggregator and podcast client
# Copyright (c) 2005-2022 The gPodder Team
#
# gPodder is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# gPodder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# libgpod_ctypes: Minimalistic ctypes-based bindings for libgpod
# (Just enough coverage to get podcast syncing working again...)
# Thomas Perl <m@thp.io>, May 2022
#


import ctypes
import logging
import os

logger = logging.getLogger(__name__)


# libgpod, for iTunesDB access
libgpod = ctypes.CDLL('libgpod.so.4')

# glib, for g_strdup() and g_free()
libglib = ctypes.CDLL('libglib-2.0.so.0')


# glib/gtypes.h: typedef gint   gboolean;
gboolean = ctypes.c_int

# glib/gstrfuncs.h: gchar *g_strdup(const gchar *str);
libglib.g_strdup.argtypes = (ctypes.c_char_p,)
# Note: This MUST be c_void_p, so that the glib-allocated buffer will
# be preserved when assigning to track member variables. The reason
# for this is that Python ctypes tries to be helpful and converts a
# c_char_p restype to a Python bytes object, which will be different
# from the memory returned by g_strdup(). For track properties, the
# values will be free'd indirectly by itdb_free() later.
libglib.g_strdup.restype = ctypes.c_void_p

# glib/gmem.h: void g_free(gpointer mem);
libglib.g_free.argtypes = (ctypes.c_void_p,)
libglib.g_free.restype = None

# ctypes.c_time_t will be available in Python 3.12 onwards
# See also: https://github.com/python/cpython/pull/92870
if hasattr(ctypes, 'c_time_t'):
    time_t = ctypes.c_time_t
else:
    # See also: https://github.com/python/cpython/issues/92869
    if ctypes.sizeof(ctypes.c_void_p) == ctypes.sizeof(ctypes.c_int64):
        time_t = ctypes.c_int64
    else:
        # On 32-bit systems, time_t is historically 32-bit, but due to Y2K38
        # there have been efforts to establish 64-bit time_t on 32-bit Linux:
        # https://linux.slashdot.org/story/20/02/15/0247201/linux-is-ready-for-the-end-of-time
        # https://www.gnu.org/software/libc/manual/html_node/64_002dbit-time-symbol-handling.html
        logger.info('libgpod may cause issues if time_t is 64-bit on your 32-bit system.')
        time_t = ctypes.c_int32


# glib/glist.h: struct _GList
class GList(ctypes.Structure):
    ...


GList._fields_ = [
    ('data', ctypes.c_void_p),
    ('next', ctypes.POINTER(GList)),
    ('prev', ctypes.POINTER(GList)),
]


# gpod/itdb.h
class Itdb_iTunesDB(ctypes.Structure):
    _fields_ = [
        ('tracks', ctypes.POINTER(GList)),
        # ...
    ]


# gpod/itdb.h: struct _Itdb_Playlist
class Itdb_Playlist(ctypes.Structure):
    _fields_ = [
        ('itdb', ctypes.POINTER(Itdb_iTunesDB)),
        ('name', ctypes.c_char_p),
        ('type', ctypes.c_uint8),
        ('flag1', ctypes.c_uint8),
        ('flag2', ctypes.c_uint8),
        ('flag3', ctypes.c_uint8),
        ('num', ctypes.c_int),
        ('members', ctypes.POINTER(GList)),
        # ...
    ]


# gpod/itdb.h
class Itdb_Chapterdata(ctypes.Structure):
    ...


# gpod/itdb.h
class Itdb_Track(ctypes.Structure):
    _fields_ = [
        ('itdb', ctypes.POINTER(Itdb_iTunesDB)),
        ('title', ctypes.c_char_p),
        ('ipod_path', ctypes.c_char_p),
        ('album', ctypes.c_char_p),
        ('artist', ctypes.c_char_p),
        ('genre', ctypes.c_char_p),
        ('filetype', ctypes.c_char_p),
        ('comment', ctypes.c_char_p),
        ('category', ctypes.c_char_p),
        ('composer', ctypes.c_char_p),
        ('grouping', ctypes.c_char_p),
        ('description', ctypes.c_char_p),
        ('podcasturl', ctypes.c_char_p),
        ('podcastrss', ctypes.c_char_p),
        ('chapterdata', ctypes.POINTER(Itdb_Chapterdata)),
        ('subtitle', ctypes.c_char_p),
        ('tvshow', ctypes.c_char_p),
        ('tvepisode', ctypes.c_char_p),
        ('tvnetwork', ctypes.c_char_p),
        ('albumartist', ctypes.c_char_p),
        ('keywords', ctypes.c_char_p),
        ('sort_artist', ctypes.c_char_p),
        ('sort_title', ctypes.c_char_p),
        ('sort_album', ctypes.c_char_p),
        ('sort_albumartist', ctypes.c_char_p),
        ('sort_composer', ctypes.c_char_p),
        ('sort_tvshow', ctypes.c_char_p),
        ('id', ctypes.c_uint32),
        ('size', ctypes.c_uint32),
        ('tracklen', ctypes.c_int32),
        ('cd_nr', ctypes.c_int32),
        ('cds', ctypes.c_int32),
        ('track_nr', ctypes.c_int32),
        ('bitrate', ctypes.c_int32),
        ('samplerate', ctypes.c_uint16),
        ('samplerate_low', ctypes.c_uint16),
        ('year', ctypes.c_int32),
        ('volume', ctypes.c_int32),
        ('soundcheck', ctypes.c_uint32),
        ('soundcheck', ctypes.c_uint32),
        ('time_added', time_t),
        ('time_modified', time_t),
        ('time_played', time_t),
        ('bookmark_time', ctypes.c_uint32),
        ('rating', ctypes.c_uint32),
        ('playcount', ctypes.c_uint32),
        ('playcount2', ctypes.c_uint32),
        ('recent_playcount', ctypes.c_uint32),
        ('transferred', gboolean),
        ('BPM', ctypes.c_int16),
        ('app_rating', ctypes.c_uint8),
        ('type1', ctypes.c_uint8),
        ('type2', ctypes.c_uint8),
        ('compilation', ctypes.c_uint8),
        ('starttime', ctypes.c_uint32),
        ('stoptime', ctypes.c_uint32),
        ('checked', ctypes.c_uint8),
        ('dbid', ctypes.c_uint64),
        ('drm_userid', ctypes.c_uint32),
        ('visible', ctypes.c_uint32),
        ('filetype_marker', ctypes.c_uint32),
        ('artwork_count', ctypes.c_uint16),
        ('artwork_size', ctypes.c_uint32),
        ('samplerate2', ctypes.c_float),
        ('unk126', ctypes.c_uint16),
        ('unk132', ctypes.c_uint32),
        ('time_released', time_t),
        ('unk144', ctypes.c_uint16),
        ('explicit_flag', ctypes.c_uint16),
        ('unk148', ctypes.c_uint32),
        ('unk152', ctypes.c_uint32),
        ('skipcount', ctypes.c_uint32),
        ('recent_skipcount', ctypes.c_uint32),
        ('last_skipped', ctypes.c_uint32),
        ('has_artwork', ctypes.c_uint8),
        ('skip_when_shuffling', ctypes.c_uint8),
        ('remember_playback_position', ctypes.c_uint8),
        ('flag4', ctypes.c_uint8),
        ('dbid2', ctypes.c_uint64),
        ('lyrics_flag', ctypes.c_uint8),
        ('movie_flag', ctypes.c_uint8),
        ('mark_unplayed', ctypes.c_uint8),
        ('unk179', ctypes.c_uint8),
        ('unk180', ctypes.c_uint32),
        ('pregap', ctypes.c_uint32),
        ('samplecount', ctypes.c_uint64),
        ('unk196', ctypes.c_uint32),
        ('postgap', ctypes.c_uint32),
        ('unk204', ctypes.c_uint32),
        ('mediatype', ctypes.c_uint32),
        # ...
    ]


# gpod/itdb.h: Itdb_iTunesDB *itdb_parse (const gchar *mp, GError **error);
libgpod.itdb_parse.argtypes = (ctypes.c_char_p, ctypes.c_void_p)
libgpod.itdb_parse.restype = ctypes.POINTER(Itdb_iTunesDB)

# gpod/itdb.h: Itdb_Playlist *itdb_playlist_podcasts (Itdb_iTunesDB *itdb);
libgpod.itdb_playlist_podcasts.argtypes = (ctypes.POINTER(Itdb_iTunesDB),)
libgpod.itdb_playlist_podcasts.restype = ctypes.POINTER(Itdb_Playlist)

# gpod/itdb.h: Itdb_Playlist *itdb_playlist_mpl (Itdb_iTunesDB *itdb);
libgpod.itdb_playlist_mpl.argtypes = (ctypes.POINTER(Itdb_iTunesDB),)
libgpod.itdb_playlist_mpl.restype = ctypes.POINTER(Itdb_Playlist)

# gpod/itdb.h: gboolean itdb_write (Itdb_iTunesDB *itdb, GError **error);
libgpod.itdb_write.argtypes = (ctypes.POINTER(Itdb_iTunesDB), ctypes.c_void_p)
libgpod.itdb_write.restype = gboolean

# gpod/itdb.h: guint32 itdb_playlist_tracks_number (Itdb_Playlist *pl);
libgpod.itdb_playlist_tracks_number.argtypes = (ctypes.POINTER(Itdb_Playlist),)
libgpod.itdb_playlist_tracks_number.restype = ctypes.c_uint32

# gpod/itdb.h: gchar *itdb_filename_on_ipod (Itdb_Track *track);
libgpod.itdb_filename_on_ipod.argtypes = (ctypes.POINTER(Itdb_Track),)
# Needs to be c_void_p, because the returned pointer-to-memory must be free'd with g_free() after use.
libgpod.itdb_filename_on_ipod.restype = ctypes.c_void_p

# gpod/itdb.h: Itdb_Track *itdb_track_new (void);
libgpod.itdb_track_new.argtypes = ()
libgpod.itdb_track_new.restype = ctypes.POINTER(Itdb_Track)

# gpod/itdb.h: void itdb_track_add (Itdb_iTunesDB *itdb, Itdb_Track *track, gint32 pos);
libgpod.itdb_track_add.argtypes = (ctypes.POINTER(Itdb_iTunesDB), ctypes.POINTER(Itdb_Track), ctypes.c_int32)
libgpod.itdb_track_add.restype = None

# gpod/itdb.h: void itdb_playlist_add_track (Itdb_Playlist *pl, Itdb_Track *track, gint32 pos);
libgpod.itdb_playlist_add_track.argtypes = (ctypes.POINTER(Itdb_Playlist), ctypes.POINTER(Itdb_Track), ctypes.c_int32)
libgpod.itdb_playlist_add_track.restype = None

# gpod/itdb.h: gboolean itdb_cp_track_to_ipod (Itdb_Track *track, const gchar *filename, GError **error);
libgpod.itdb_cp_track_to_ipod.argtypes = (ctypes.POINTER(Itdb_Track), ctypes.c_char_p, ctypes.c_void_p)
libgpod.itdb_cp_track_to_ipod.restype = gboolean

# gpod/itdb.h: time_t itdb_time_host_to_mac (time_t time);
libgpod.itdb_time_host_to_mac.argtypes = (time_t,)
libgpod.itdb_time_host_to_mac.restype = time_t

# gpod/itdb.h: void itdb_playlist_remove_track (Itdb_Playlist *pl, Itdb_Track *track);
libgpod.itdb_playlist_remove_track.argtypes = (ctypes.POINTER(Itdb_Playlist), ctypes.POINTER(Itdb_Track))
libgpod.itdb_playlist_remove_track.restype = None

# gpod/itdb.h: void itdb_track_remove (Itdb_Track *track);
libgpod.itdb_track_remove.argtypes = (ctypes.POINTER(Itdb_Track),)
libgpod.itdb_track_remove.restype = None

# gpod/itdb.h: void itdb_free (Itdb_iTunesDB *itdb);
libgpod.itdb_free.argtypes = (ctypes.POINTER(Itdb_iTunesDB),)
libgpod.itdb_free.restype = None


# gpod/itdb.h
ITDB_MEDIATYPE_AUDIO = (1 << 0)
ITDB_MEDIATYPE_MOVIE = (1 << 1)
ITDB_MEDIATYPE_PODCAST = (1 << 2)
ITDB_MEDIATYPE_VIDEO_PODCAST = (ITDB_MEDIATYPE_MOVIE | ITDB_MEDIATYPE_PODCAST)


def glist_foreach(ptr_to_glist, item_type):
    cur = ptr_to_glist
    while cur:
        yield ctypes.cast(cur[0].data, item_type)
        if not cur[0].next:
            break
        cur = cur[0].next


class iPodTrack(object):
    def __init__(self, db, track):
        self.db = db
        self.track = track

        self.episode_title = track[0].title.decode()
        self.podcast_title = track[0].album.decode()

        self.podcast_url = track[0].podcasturl.decode()
        self.podcast_rss = track[0].podcastrss.decode()

        self.playcount = track[0].playcount
        self.bookmark_time = track[0].bookmark_time

        # This returns a newly-allocated string, so we have to juggle the memory
        # around a bit and take a copy of the string before free'ing it again.
        filename_ptr = libgpod.itdb_filename_on_ipod(track)
        if filename_ptr:
            self.filename_on_ipod = ctypes.string_at(filename_ptr).decode()
            libglib.g_free(filename_ptr)
        else:
            self.filename_on_ipod = None

    def __repr__(self):
        return 'iPodTrack(episode={}, podcast={})'.format(self.episode_title, self.podcast_title)

    def initialize_bookmark(self, is_new, bookmark_time):
        self.track[0].mark_unplayed = 0x02 if is_new else 0x01
        self.track[0].bookmark_time = int(bookmark_time)

    def remove_from_device(self):
        libgpod.itdb_playlist_remove_track(self.db.podcasts_playlist, self.track)
        libgpod.itdb_playlist_remove_track(self.db.master_playlist, self.track)

        # This frees the memory pointed-to by the track object
        libgpod.itdb_track_remove(self.track)

        self.track = None

        # Don't forget to write the database on close
        self.db.modified = True

        if self.filename_on_ipod is not None:
            try:
                os.unlink(self.filename_on_ipod)
            except Exception:
                logger.info('Could not delete podcast file from iPod', exc_info=True)


class iPodDatabase(object):
    def __init__(self, mountpoint):
        self.mountpoint = mountpoint
        self.itdb = libgpod.itdb_parse(mountpoint.encode(), None)

        if not self.itdb:
            raise ValueError('iTunesDB not found at {}'.format(self.mountpoint))

        logger.info('iTunesDB: %s', self.itdb)

        self.modified = False

        self.podcasts_playlist = libgpod.itdb_playlist_podcasts(self.itdb)
        self.master_playlist = libgpod.itdb_playlist_mpl(self.itdb)

        self.tracks = [iPodTrack(self, track)
                       for track in glist_foreach(self.podcasts_playlist[0].members, ctypes.POINTER(Itdb_Track))]

    def get_podcast_tracks(self):
        return self.tracks

    def add_track(self, filename, episode_title, podcast_title, description, podcast_url, podcast_rss,
            published_timestamp, track_length, is_audio):
        track = libgpod.itdb_track_new()

        track[0].title = libglib.g_strdup(episode_title.encode())
        track[0].album = libglib.g_strdup(podcast_title.encode())
        track[0].artist = libglib.g_strdup(podcast_title.encode())
        track[0].description = libglib.g_strdup(description.encode())
        track[0].podcasturl = libglib.g_strdup(podcast_url.encode())
        track[0].podcastrss = libglib.g_strdup(podcast_rss.encode())
        track[0].tracklen = track_length
        track[0].size = os.path.getsize(filename)
        track[0].time_released = libgpod.itdb_time_host_to_mac(published_timestamp)

        if is_audio:
            track[0].filetype = libglib.g_strdup(b'mp3')
            track[0].mediatype = ITDB_MEDIATYPE_PODCAST
        else:
            track[0].filetype = libglib.g_strdup(b'm4v')
            track[0].mediatype = ITDB_MEDIATYPE_VIDEO_PODCAST

        # Start at the beginning, and add "unplayed" bullet
        track[0].bookmark_time = 0
        track[0].mark_unplayed = 0x02

        # from set_podcast_flags()
        track[0].remember_playback_position = 0x01
        track[0].skip_when_shuffling = 0x01
        track[0].flag1 = 0x02
        track[0].flag2 = 0x01
        track[0].flag3 = 0x01
        track[0].flag4 = 0x01

        libgpod.itdb_track_add(self.itdb, track, -1)

        libgpod.itdb_playlist_add_track(self.podcasts_playlist, track, -1)
        libgpod.itdb_playlist_add_track(self.master_playlist, track, -1)

        copied = libgpod.itdb_cp_track_to_ipod(track, filename.encode(), None)
        logger.info('Copy result: %r', copied)
        self.modified = True

        self.tracks.append(iPodTrack(self, track))
        return self.tracks[-1]

    def __del__(self):
        # If we hit the finalizer without closing the iTunesDB properly,
        # just free the memory, but don't write out any modifications.
        self.close(write=False)

    def close(self, write=True):
        if self.itdb:
            if self.modified and write:
                result = libgpod.itdb_write(self.itdb, None)
                logger.info('Close result: %r', result)
                self.modified = False

            libgpod.itdb_free(self.itdb)
            self.itdb = None


if __name__ == '__main__':
    import argparse
    import textwrap

    parser = argparse.ArgumentParser(description='Dump podcasts in iTunesDB via libgpod')
    parser.add_argument('mountpoint', type=str, help='Path to mounted iPod storage')

    args = parser.parse_args()

    ipod = iPodDatabase(args.mountpoint)

    for track in ipod.get_podcast_tracks():
        print(textwrap.dedent(f"""
        Episode:     {track.episode_title}
        Podcast:     {track.podcast_title}
        Episode URL: {track.podcast_url}
        Podcast URL: {track.podcast_rss}
        Play count:  {track.playcount}
        Bookmark:    {track.bookmark_time / 1000:.0f} seconds
        Filename:    {track.filename_on_ipod}
        """).rstrip())

    ipod.close()



# -*- coding: utf-8 -*-
#
# gPodder - A media aggregator and podcast client
# Copyright (c) 2005-2018 The gPodder Team
#
# gPodder is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# gPodder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


# sync.py -- Device synchronization
# Thomas Perl <thp@perli.net> 2007-12-06
# based on libipodsync.py (2006-04-05 Thomas Perl)
# Ported to gPodder 3 by Joseph Wickremasinghe in June 2012

import logging
import os.path
import threading
import time

import gpodder
from gpodder import download, services, util

import gi  # isort:skip
gi.require_version('Gio', '2.0')  # isort:skip
from gi.repository import GLib, Gio  # isort:skip


logger = logging.getLogger(__name__)


_ = gpodder.gettext

gpod_available = True
try:
    from gpodder import libgpod_ctypes
except:
    logger.info('iPod sync not available')
    gpod_available = False

mplayer_available = True if util.find_command('mplayer') is not None else False

eyed3mp3_available = True
try:
    import eyed3.mp3
except:
    logger.info('eyeD3 MP3 not available')
    eyed3mp3_available = False


def open_device(gui):
    config = gui._config
    device_type = gui._config.device_sync.device_type
    if device_type == 'ipod':
        return iPodDevice(config,
                gui.download_status_model,
                gui.download_queue_manager)
    elif device_type == 'filesystem':
        return MP3PlayerDevice(config,
                gui.download_status_model,
                gui.download_queue_manager,
                gui.mount_volume_for_file)

    return None


def get_track_length(filename):
    attempted = False

    if mplayer_available:
        try:
            mplayer_output = os.popen('mplayer -msglevel all=-1 -identify -vo null -ao null -frames 0 "%s" 2>/dev/null' % filename).read()
            return int(float(mplayer_output[mplayer_output.index('ID_LENGTH'):].splitlines()[0][10:]) * 1000)
        except Exception:
            logger.error('MPlayer could not determine length: %s', filename, exc_info=True)
            attempted = True

    if eyed3mp3_available:
        try:
            length = int(eyed3.mp3.Mp3AudioFile(filename).info.time_secs * 1000)
            # Notify user on eyed3 success if mplayer failed.
            # A warning is used to make it visible in gpo or on console.
            if attempted:
                logger.warning('eyed3.mp3 successfully determined length: %s', filename)
            return length
        except Exception:
            logger.error('eyed3.mp3 could not determine length: %s', filename, exc_info=True)
            attempted = True

    if not attempted:
        logger.warning('Could not determine length: %s', filename)
        logger.warning('Please install MPlayer or the eyed3.mp3 module for track length detection.')

    return int(60 * 60 * 1000 * 3)
    # Default is three hours (to be on the safe side)


def episode_filename_on_device(config, episode):
    """Return the basename of the episode file to save on device.

    :param gpodder.config.Config config: configuration (for sync options)
    :param gpodder.model.PodcastEpisode episode: episode to get filename for
    :return str: basename minus extension to use to save episode on device
    """
    # get the local file
    from_file = episode.local_filename(create=False)
    # get the formatted base name
    filename_base = util.sanitize_filename(episode.sync_filename(
        config.device_sync.custom_sync_name_enabled,
        config.device_sync.custom_sync_name),
        config.device_sync.max_filename_length)
    # add the file extension
    to_file = filename_base + os.path.splitext(from_file)[1].lower()

    # dirty workaround: on bad (empty) episode titles,
    # we simply use the from_file basename
    # (please, podcast authors, FIX YOUR RSS FEEDS!)
    if os.path.splitext(to_file)[0] == '':
        to_file = os.path.basename(from_file)
    return to_file


def episode_foldername_on_device(config, episode):
    """Return the folder name to which the episode is saved on device.

    :param gpodder.config.Config config: configuration (for sync options)
    :param gpodder.model.PodcastEpisode episode: episode to get folder name for
    :return str: folder name to save episode to on device
    """
    if config.device_sync.one_folder_per_podcast:
        # Add channel title as subfolder
        folder = episode.channel.title
        # Clean up the folder name for use on limited devices
        folder = util.sanitize_filename(folder, config.device_sync.max_filename_length)
    else:
        folder = None
    return folder


class SyncTrack(object):
    """Class representing a track that is on a device.

    You need to specify at least the following keyword arguments,
    because these will be used to display the track in the
    GUI. All other keyword arguments are optional and can
    be used to reference internal objects, etc... See the
    iPod synchronization code for examples.

    Keyword arguments needed:
        playcount (How often has the track been played?)
        podcast (Which podcast is this track from? Or: Folder name)

    If any of these fields is unknown, it should not be
    passed to the function (the values will default to None
    for all required fields).
    """

    def __init__(self, title, length, modified, **kwargs):
        self.title = title
        self.length = length
        self.filesize = util.format_filesize(length)
        self.modified = modified

        # Set some (possible) keyword arguments to default values
        self.playcount = 0
        self.podcast = None

        # Convert keyword arguments to object attributes
        self.__dict__.update(kwargs)

    def __repr__(self):
        return 'SyncTrack(title={}, podcast={})'.format(self.title, self.podcast)

    @property
    def playcount_str(self):
        return str(self.playcount)


class Device(services.ObservableService):
    def __init__(self, config):
        self._config = config
        self.cancelled = False
        self.allowed_types = ['audio', 'video']
        self.errors = []
        self.tracks_list = []
        signals = ['progress', 'sub-progress', 'status', 'done', 'post-done']
        services.ObservableService.__init__(self, signals)

    def open(self):
        pass

    def cancel(self):
        self.cancelled = True
        self.notify('status', _('Cancelled by user'))

    def close(self):
        self.notify('status', _('Writing data to disk'))
        if self._config.device_sync.after_sync.sync_disks and not gpodder.ui.win32:
            os.system('sync')
        else:
            logger.warning('Not syncing disks. Unmount your device before unplugging.')
        return True

    def create_task(self, track):
        return SyncTask(track)

    def cancel_task(self, task):
        pass

    def cleanup_task(self, task):
        pass

    def add_sync_tasks(self, tracklist, force_played=False, done_callback=None):
        for track in list(tracklist):
            # Filter tracks that are not meant to be synchronized
            does_not_exist = not track.was_downloaded(and_exists=True)
            exclude_played = (not track.is_new
                    and self._config.device_sync.skip_played_episodes)
            wrong_type = track.file_type() not in self.allowed_types

            if does_not_exist:
                tracklist.remove(track)
            elif exclude_played or wrong_type:
                logger.info('Excluding %s from sync', track.title)
                tracklist.remove(track)

        if tracklist:
            for track in sorted(tracklist, key=lambda e: e.pubdate_prop):
                if self.cancelled:
                    break

                # XXX: need to check if track is added properly?
                sync_task = self.create_task(track)

                sync_task.status = sync_task.NEW
                sync_task.device = self
                # New Task, we must wait on the GTK Loop
                self.download_status_model.register_task(sync_task)
                # Executes after task has been registered
                util.idle_add(self.download_queue_manager.queue_task, sync_task)
        else:
            logger.warning("No episodes to sync")

        if done_callback:
            done_callback()

    def get_all_tracks(self):
        pass

    def add_track(self, track, reporthook=None):
        pass

    def remove_track(self, track):
        pass

    def get_free_space(self):
        pass

    def episode_on_device(self, episode):
        return self._track_on_device(episode.title)

    def _track_on_device(self, track_name):
        for t in self.tracks_list:
            title = t.title
            if track_name == title:
                return t
        return None


class iPodDevice(Device):
    def __init__(self, config,
            download_status_model,
            download_queue_manager):
        Device.__init__(self, config)

        self.mountpoint = self._config.device_sync.device_folder
        self.download_status_model = download_status_model
        self.download_queue_manager = download_queue_manager

        self.ipod = None

    def get_free_space(self):
        # Reserve 10 MiB for iTunesDB writing (to be on the safe side)
        RESERVED_FOR_ITDB = 1024 * 1024 * 10
        result = util.get_free_disk_space(self.mountpoint)
        if result == -1:
            # Can't get free disk space
            return -1
        return result - RESERVED_FOR_ITDB

    def open(self):
        Device.open(self)
        if not gpod_available:
            logger.error('Please install libgpod 0.8.3 to sync with an iPod device.')
            return False
        if not os.path.isdir(self.mountpoint):
            return False

        self.notify('status', _('Opening iPod database'))
        self.ipod = libgpod_ctypes.iPodDatabase(self.mountpoint)

        if not self.ipod.itdb or not self.ipod.podcasts_playlist or not self.ipod.master_playlist:
            return False

        self.notify('status', _('iPod opened'))

        # build the initial tracks_list
        self.tracks_list = self.get_all_tracks()

        return True

    def close(self):
        if self.ipod is not None:
            self.notify('status', _('Saving iPod database'))
            self.ipod.close()
            self.ipod = None

        Device.close(self)
        return True

    def get_all_tracks(self):
        tracks = []
        for track in self.ipod.get_podcast_tracks():
            filename = track.filename_on_ipod

            if filename is None:
                length = 0
                modified = ''
            else:
                length = util.calculate_size(filename)
                timestamp = util.file_modification_timestamp(filename)
                modified = util.format_date(timestamp)

            t = SyncTrack(track.episode_title, length, modified,
                    ipod_track=track,
                    playcount=track.playcount,
                    podcast=track.podcast_title)
            tracks.append(t)
        return tracks

    def episode_on_device(self, episode):
        return next((track for track in self.tracks_list
                     if track.ipod_track.podcast_rss == episode.channel.url
                     and track.ipod_track.podcast_url == episode.url), None)

    def remove_track(self, track):
        self.notify('status', _('Removing %s') % track.title)
        logger.info('Removing track from iPod: %r', track.title)
        track.ipod_track.remove_from_device()
        try:
            self.tracks_list.remove(next((sync_track for sync_track in self.tracks_list
                                          if sync_track.ipod_track == track), None))
        except ValueError:
            ...

    def add_track(self, task, reporthook=None):
        episode = task.episode
        self.notify('status', _('Adding %s') % episode.title)
        tracklist = self.ipod.get_podcast_tracks()
        episode_urls = [track.podcast_url for track in tracklist]

        if episode.url in episode_urls:
            # Mark as played on iPod if played locally (and set podcast flags)
            self.update_from_episode(tracklist[episode_urls.index(episode.url)], episode)
            return True

        local_filename = episode.local_filename(create=False)
        # The file has to exist, if we ought to transfer it, and therefore,
        # local_filename(create=False) must never return None as filename
        assert local_filename is not None

        if util.calculate_size(local_filename) > self.get_free_space():
            logger.error('Not enough space on %s, sync aborted...', self.mountpoint)
            d = {'episode': episode.title, 'mountpoint': self.mountpoint}
            message = _('Error copying %(episode)s: Not enough free space on %(mountpoint)s')
            self.errors.append(message % d)
            self.cancelled = True
            return False

        (fn, extension) = os.path.splitext(local_filename)
        if extension.lower().endswith('ogg'):
            # XXX: Proper file extension/format support check for iPod
            logger.error('Cannot copy .ogg files to iPod.')
            return False

        track = self.ipod.add_track(local_filename, episode.title, episode.channel.title,
                episode._text_description, episode.url, episode.channel.url,
                episode.published, get_track_length(local_filename), episode.file_type() == 'audio')

        self.update_from_episode(track, episode, initial=True)

        reporthook(episode.file_size, 1, episode.file_size)

        return True

    def update_from_episode(self, track, episode, *, initial=False):
        if initial:
            # Set the initial bookmark on the device based on what we have locally
            track.initialize_bookmark(episode.is_new, episode.current_position * 1000)
        else:
            # Copy updated status from iPod
            if track.playcount > 0:
                episode.is_new = False

            if track.bookmark_time > 0:
                logger.info('Playback position from iPod: %s', util.format_time(track.bookmark_time / 1000))
                episode.is_new = False
                episode.current_position = int(track.bookmark_time / 1000)
                episode.current_position_updated = time.time()

            episode.save()


class MP3PlayerDevice(Device):
    def __init__(self, config,
            download_status_model,
            download_queue_manager,
            mount_volume_for_file):
        Device.__init__(self, config)

        folder = self._config.device_sync.device_folder
        self.destination = util.new_gio_file(folder)
        self.mount_volume_for_file = mount_volume_for_file
        self.download_status_model = download_status_model
        self.download_queue_manager = download_queue_manager

    def get_free_space(self):
        info = self.destination.query_filesystem_info(Gio.FILE_ATTRIBUTE_FILESYSTEM_FREE, None)
        return info.get_attribute_uint64(Gio.FILE_ATTRIBUTE_FILESYSTEM_FREE)

    def open(self):
        Device.open(self)
        self.notify('status', _('Opening MP3 player'))

        if not self.mount_volume_for_file(self.destination):
            return False

        try:
            info = self.destination.query_info(
                Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE + ","
                + Gio.FILE_ATTRIBUTE_STANDARD_TYPE,
                Gio.FileQueryInfoFlags.NONE,
                None)
        except GLib.Error as err:
            logger.error('querying destination info for %s failed with %s',
                self.destination.get_uri(), err.message)
            return False

        if info.get_file_type() != Gio.FileType.DIRECTORY:
            logger.error('destination %s is not a directory', self.destination.get_uri())
            return False

        # open is ok if the target is a directory, and it can be written to
        # for smb, query_info doesn't return FILE_ATTRIBUTE_ACCESS_CAN_WRITE,
        # -- if that's the case, just assume that it's writable
        if (not info.has_attribute(Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE)
                or info.get_attribute_boolean(Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE)):
            self.notify('status', _('MP3 player opened'))
            self.tracks_list = self.get_all_tracks()
            return True

        logger.error('destination %s is not writable', self.destination.get_uri())
        return False

    def get_episode_folder_on_device(self, episode):
        folder = episode_foldername_on_device(self._config, episode)
        if folder:
            folder = self.destination.get_child(folder)
        else:
            folder = self.destination

        return folder

    def get_episode_file_on_device(self, episode):
        return episode_filename_on_device(self._config, episode)

    def create_task(self, track):
        return GioSyncTask(track)

    def cancel_task(self, task):
        task.cancellable.cancel()

    # called by the sync task when it is removed and needs partial files cleaning up
    def cleanup_task(self, task):
        episode = task.episode
        folder = self.get_episode_folder_on_device(episode)
        file = self.get_episode_file_on_device(episode)
        file = folder.get_child(file)
        self.remove_track_file(file)

    def add_track(self, task, reporthook=None):
        episode = task.episode
        self.notify('status', _('Adding %s') % episode.title)

        # get the folder on the device
        folder = self.get_episode_folder_on_device(episode)

        filename = episode.local_filename(create=False)
        # The file has to exist, if we ought to transfer it, and therefore,
        # local_filename(create=False) must never return None as filename
        assert filename is not None

        from_file = filename

        # verify free space
        needed = util.calculate_size(from_file)
        free = self.get_free_space()
        if free == -1:
            logger.warning('Cannot determine free disk space on device')
        elif needed > free:
            d = {'path': self.destination, 'free': util.format_filesize(free), 'need': util.format_filesize(needed)}
            message = _('Not enough space in %(path)s: %(free)s available, but need at least %(need)s')
            raise SyncFailedException(message % d)

        # get the filename that will be used on the device
        to_file = self.get_episode_file_on_device(episode)
        to_file = folder.get_child(to_file)

        util.make_directory(folder)

        to_file_exists = to_file.query_exists()
        from_size = episode.file_size
        to_size = episode.file_size
        # An interrupted sync results in a partial file on the device that must be removed to fully sync it.
        # Comparing file size would detect such files and finish uploading.
        # However, some devices add metadata to files, increasing their size, and forcing an upload on every sync.
        # File size and checksum can not be used.
        if to_file_exists and self._config.device_sync.compare_episode_filesize:
            try:
                info = to_file.query_info(Gio.FILE_ATTRIBUTE_STANDARD_SIZE, Gio.FileQueryInfoFlags.NONE)
                to_size = info.get_attribute_uint64(Gio.FILE_ATTRIBUTE_STANDARD_SIZE)
            except GLib.Error:
                # Assume same size and don't sync again
                pass
        if not to_file_exists or from_size != to_size:
            logger.info('Copying %s (%d bytes) => %s (%d bytes)',
                    os.path.basename(from_file), from_size,
                    to_file.get_uri(), to_size)
            from_file = Gio.File.new_for_path(from_file)
            try:
                def hookconvert(current_bytes, total_bytes, user_data):
                    return reporthook(current_bytes, 1, total_bytes)
                from_file.copy(to_file, Gio.FileCopyFlags.OVERWRITE, task.cancellable, hookconvert, None)
            except GLib.Error as err:
                if err.matches(Gio.io_error_quark(), Gio.IOErrorEnum.CANCELLED):
                    raise SyncCancelledException()
                logger.error('Error copying %s to %s: %s', from_file.get_uri(), to_file.get_uri(), err.message)
                d = {'from_file': from_file.get_uri(), 'to_file': to_file.get_uri(), 'message': err.message}
                self.errors.append(_('Error copying %(from_file)s to %(to_file)s: %(message)s') % d)
                return False

        return True

    def add_sync_track(self, tracks, file, info, podcast_name):
        (title, extension) = os.path.splitext(info.get_name())
        timestamp = info.get_modification_time()
        modified = util.format_date(timestamp.tv_sec)

        t = SyncTrack(title, info.get_size(), modified,
                filename=file.get_uri(),
                podcast=podcast_name)
        tracks.append(t)

    def get_all_tracks(self):
        tracks = []

        attributes = (
            Gio.FILE_ATTRIBUTE_STANDARD_NAME + ","
            + Gio.FILE_ATTRIBUTE_STANDARD_TYPE + ","
            + Gio.FILE_ATTRIBUTE_STANDARD_SIZE + ","
            + Gio.FILE_ATTRIBUTE_TIME_MODIFIED)

        root_path = self.destination
        for path_info in root_path.enumerate_children(attributes, Gio.FileQueryInfoFlags.NONE, None):
            if self._config.one_folder_per_podcast:
                if path_info.get_file_type() == Gio.FileType.DIRECTORY:
                    path_file = root_path.get_child(path_info.get_name())
                    try:
                        for child_info in path_file.enumerate_children(attributes, Gio.FileQueryInfoFlags.NONE, None):
                            if child_info.get_file_type() == Gio.FileType.REGULAR:
                                child_file = path_file.get_child(child_info.get_name())
                                self.add_sync_track(tracks, child_file, child_info, path_info.get_name())
                    except GLib.Error as err:
                        logger.error('get all tracks for %s failed: %s', path_file.get_uri(), err.message)

            else:
                if path_info.get_file_type() == Gio.FileTypeFlags.REGULAR:
                    path_file = root_path.get_child(path_info.get_name())
                    self.add_sync_track(tracks, path_file, path_info, None)
        return tracks

    def episode_on_device(self, episode):
        e = util.sanitize_filename(episode.sync_filename(
            self._config.device_sync.custom_sync_name_enabled,
            self._config.device_sync.custom_sync_name),
            self._config.device_sync.max_filename_length)
        return self._track_on_device(e)

    def remove_track_file(self, file):
        folder = file.get_parent()
        if file.query_exists():
            try:
                file.delete()
            except GLib.Error as err:
                # if the file went away don't worry about it
                if not err.matches(Gio.io_error_quark(), Gio.IOErrorEnum.NOT_FOUND):
                    logger.error('deleting file %s failed: %s', file.get_uri(), err.message)
                return

        if self._config.one_folder_per_podcast:
            try:
                if self.directory_is_empty(folder):
                    folder.delete()
            except GLib.Error as err:
                # if the folder went away don't worry about it (multiple threads could
                # make this happen if they both notice the folder is empty simultaneously)
                if not err.matches(Gio.io_error_quark(), Gio.IOErrorEnum.NOT_FOUND):
                    logger.error('deleting folder %s failed: %s', folder.get_uri(), err.message)

    def remove_track(self, track):
        self.notify('status', _('Removing %s') % track.title)

        # get the folder on the device
        file = Gio.File.new_for_uri(track.filename)
        self.remove_track_file(file)

    def directory_is_empty(self, directory):
        for child in directory.enumerate_children(Gio.FILE_ATTRIBUTE_STANDARD_NAME, Gio.FileQueryInfoFlags.NONE, None):
            return False
        return True


class SyncCancelledException(Exception):
    pass


class SyncFailedException(Exception):
    pass


class SyncTask(download.DownloadTask):
    # An object representing the synchronization task of an episode

    # Possible states this sync task can be in
    STATUS_MESSAGE = (_('Queued'), _('Queued'), _('Syncing'),
            _('Finished'), _('Failed'), _('Cancelling'), _('Cancelled'), _('Pausing'), _('Paused'))
    (NEW, QUEUED, DOWNLOADING, DONE, FAILED, CANCELLING, CANCELLED, PAUSING, PAUSED) = list(range(9))

    def __str__(self):
        return self.__episode.title

    def __get_status(self):
        return self.__status

    def __set_status(self, status):
        if status != self.__status:
            self.__status_changed = True
            self.__status = status

    status = property(fget=__get_status, fset=__set_status)

    def __get_device(self):
        return self.__device

    def __set_device(self, device):
        self.__device = device

    device = property(fget=__get_device, fset=__set_device)

    def __get_status_changed(self):
        if self.__status_changed:
            self.__status_changed = False
            return True
        else:
            return False

    status_changed = property(fget=__get_status_changed)

    def __get_activity(self):
        return self.__activity

    def __set_activity(self, activity):
        self.__activity = activity

    activity = property(fget=__get_activity, fset=__set_activity)

    def __get_empty_string(self):
        return ''

    url = property(fget=__get_empty_string)
    podcast_url = property(fget=__get_empty_string)

    def __get_episode(self):
        return self.__episode

    episode = property(fget=__get_episode)

    def can_queue(self):
        return self.status in (self.CANCELLED, self.PAUSED, self.FAILED)

    def can_pause(self):
        return self.status in (self.DOWNLOADING, self.QUEUED)

    def pause(self):
        with self:
            # Pause a queued download
            if self.status == self.QUEUED:
                self.status = self.PAUSED
            # Request pause of a running download
            elif self.status == self.DOWNLOADING:
                self.status = self.PAUSING

    def can_cancel(self):
        return self.status in (self.DOWNLOADING, self.QUEUED, self.PAUSED, self.FAILED)

    def cancel(self):
        with self:
            # Cancelling directly is allowed if the task isn't currently downloading
            if self.status in (self.QUEUED, self.PAUSED, self.FAILED):
                self.status = self.CANCELLED
                # Call run, so the partial file gets deleted
                self.run()
                self.recycle()
            # Otherwise request cancellation
            elif self.status == self.DOWNLOADING:
                self.status = self.CANCELLING
                self.device.cancel()

    def can_remove(self):
        return self.status in (self.CANCELLED, self.FAILED, self.DONE)

    def removed_from_list(self):
        if self.status != self.DONE:
            self.device.cleanup_task(self)

    def __init__(self, episode):
        self.__lock = threading.RLock()
        self.__status = SyncTask.NEW
        self.__activity = SyncTask.ACTIVITY_SYNCHRONIZE
        self.__status_changed = True
        self.__episode = episode

        # Create the target filename and save it in the database
        self.filename = self.__episode.local_filename(create=False)

        self.total_size = self.__episode.file_size
        self.speed = 0.0
        self.progress = 0.0
        self.error_message = None
        self.custom_downloader = None

        # Have we already shown this task in a notification?
        self._notification_shown = False

        # Variables for speed limit and speed calculation
        self.__start_time = 0
        self.__start_blocks = 0
        self.__limit_rate_value = 999
        self.__limit_rate = 999

        # Callbacks
        self._progress_updated = lambda x: None

    def __enter__(self):
        return self.__lock.acquire()

    def __exit__(self, exception_type, value, traceback):
        self.__lock.release()

    def notify_as_finished(self):
        if self.status == SyncTask.DONE:
            if self._notification_shown:
                return False
            else:
                self._notification_shown = True
                return True

        return False

    def notify_as_failed(self):
        if self.status == SyncTask.FAILED:
            if self._notification_shown:
                return False
            else:
                self._notification_shown = True
                return True

        return False

    def add_progress_callback(self, callback):
        self._progress_updated = callback

    def status_updated(self, count, blockSize, totalSize):
        # We see a different "total size" while downloading,
        # so correct the total size variable in the thread
        if totalSize != self.total_size and totalSize > 0:
            self.total_size = float(totalSize)

        if self.total_size > 0:
            self.progress = max(0.0, min(1.0, (count * blockSize) / self.total_size))
            self._progress_updated(self.progress)

        if self.status in (SyncTask.CANCELLING, SyncTask.PAUSING):
            self._signal_cancel_from_status()

    # default implementation
    def _signal_cancel_from_status(self):
        raise SyncCancelledException()

    def recycle(self):
        self.episode.download_task = None

    def run(self):
        # Speed calculation (re-)starts here
        self.__start_time = 0
        self.__start_blocks = 0

        # If the download has already been cancelled/paused, skip it
        with self:
            if self.status in (SyncTask.CANCELLING, SyncTask.CANCELLED):
                self.progress = 0.0
                self.speed = 0.0
                self.status = SyncTask.CANCELLED
                return False

            if self.status == SyncTask.PAUSING:
                self.status = SyncTask.PAUSED
                return False

            # We only start this download if its status is downloading
            if self.status != SyncTask.DOWNLOADING:
                return False

            # We are syncing this file right now
            self._notification_shown = False

        sync_result = SyncTask.DOWNLOADING
        try:
            logger.info('Starting SyncTask')
            self.device.add_track(self, reporthook=self.status_updated)
        except SyncCancelledException:
            sync_result = SyncTask.CANCELLED
        except Exception as e:
            sync_result = SyncTask.FAILED
            logger.error('Sync failed: %s', str(e), exc_info=True)
            self.error_message = _('Error: %s') % (str(e),)

        with self:
            if sync_result == SyncTask.DOWNLOADING:
                # Everything went well - we're done
                self.status = SyncTask.DONE
                if self.total_size <= 0:
                    self.total_size = util.calculate_size(self.filename)
                    logger.info('Total size updated to %d', self.total_size)
                self.progress = 1.0
                gpodder.user_extensions.on_episode_synced(self.device, self.__episode)
                return True

            self.speed = 0.0

            if sync_result == SyncTask.FAILED:
                self.status = SyncTask.FAILED

            # cancelled/paused -- update state to mark it as safe to manipulate this task again
            elif self.status == SyncTask.PAUSING:
                self.status = SyncTask.PAUSED
            elif self.status == SyncTask.CANCELLING:
                self.status = SyncTask.CANCELLED

        # We finished, but not successfully (at least not really)
        return False


class GioSyncTask(SyncTask):
    def __init__(self, episode):
        super().__init__(episode)
        # For cancelling the copy
        self.cancellable = Gio.Cancellable()

    def _signal_cancel_from_status(self):
        self.cancellable.cancel()
